#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
@author: zoharslong
"""
from pyzohar.sub_slt_bsc.bsz import stz, lsz, dtz, dcz
from pyzohar.sub_slt_bsc.dfz import dfz

__all__ = ['stz', 'lsz', 'dtz', 'dcz', 'dfz']
print('info: basic slot inserted.')
