#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.05.31
A package slot for ppt generating
requirements = [
  'python-pptx>=0.6.19',
]
@author: zoharslong
"""
from pyzohar.sub_slt_ppt.ppz import ppz

__all__ = ['ppz']
print('info: ppt generating slot inserted.')
