#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
A package slot for modelling
requirements = [
    'matplotlib>=3.1.3',
    'sklearn>=0.22.1',
    'seaborn>=0.10.0',
    'factor_analyzer>=0.3.2',
    'joblib>=0.14.1',
]
@author: zoharslong
"""
from pyzohar.sub_slt_mdl.mdz import mdz

__all__ = ['mdz']
print('info: modelling slot inserted.')
