#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on 2021.03.19
A package slot for web spider
@author: zoharslong
"""
from pyzohar.sub_slt_spd.spz import spz

__all__ = ['spz']
print('info: web spider slot inserted.')
